local function deepcopy(object)
    local lookup_table = {}
    local function _copy(object)
        if type(object) ~= "table" then
            return object
        elseif lookup_table[object] then
            return lookup_table[object]
        end
        local new_table = {}
        lookup_table[object] = new_table
        for index, value in pairs(object) do
            new_table[_copy(index)] = _copy(value)
        end
        return setmetatable(new_table, getmetatable(object))
    end
    return _copy(object)
end

local function Gnomario_World_Unregister(self)
	local sprites = self.sprites;
	for n = 1, #sprites do
		local sprite = sprites[n];
		sprite.texture:Hide();
		sprite.texture:SetTexture(nil);
		sprite.id = nil;
	end
	self.world = {};
	self.creatures = {};
	self.environment = {};
end

local function Gnomario_World_Add(self)
	local sprites = self.sprites;
	for n = 1, #sprites do
		local sprite = sprites[n];
		if sprite and not sprite.id then
			sprite.id = time();
			return sprite;
		end
	end
	local sprite = {};
	sprite.texture = UIParent:CreateTexture();
	sprite.id = time();
	table.insert(self.sprites, sprite);
	return sprite;
end

local function Gnomario_World_Turn(self)
	self.speed = -self.speed;
end

local function Gnomario_World_Walk(self)
	self.forceH = self.speed;
end

function Gnomario_World_Spawn_Object(self, typ)
	if not self.xml then error("Could not find object database"); end
	local mob = nil;
	for key, obj in pairs(self.xml.environment or {}) do
		if type(obj) == "table" and obj.args and obj.args.id == typ then
			mob = {};
			mob.id = obj.args.id;
			mob.static = true;
			mob.width = obj.args.width or 20;
			mob.height = obj.args.height or 20;
			mob.tex = obj.args.tex;
			break;
		end
	end
	if mob then
		mob.sprite = Gnomario_World_Add(self);
		if mob.tex then
			mob.sprite.texture:SetTexture([[Interface\AddOns\Gnomario\Textures\]] .. mob.tex);
		else
			mob.sprite.texture:SetTexture(1,1,0,1);
		end
		mob.sprite.texture:SetSize(mob.width, mob.height);
		mob.sprite.texture:SetParent(self.frame);
		mob.sprite.obj = mob;
		tinsert(self.creatures, mob);
		self.physics:Put(mob, 0, 0);
	end
	return mob;
end

function Gnomario_World_Spawn_Creature(self, typ)
	if not self.xml then error("Could not find object database"); end
	local mob = nil;
	for key, obj in pairs(self.xml.creatures or {}) do
		if type(obj) == "table" and obj.args and obj.args.id == typ then
			mob = {};
			mob.id = obj.args.id;
			mob.width = obj.args.width or 20;
			mob.height = obj.args.height or 20;
			mob.speed = tonumber(obj.args.speed) or 1;
			mob.health = tonumber(obj.args.health) or 1;
			if obj.args.hostile == "true" then mob.hostile = true; end
			mob.tex = obj.args.tex;
			break;
		end
	end
	if mob then
		mob.sprite = Gnomario_World_Add(self);
		if mob.tex then
			mob.sprite.texture:SetTexture([[Interface\AddOns\Gnomario\Textures\]] .. mob.tex);
		else
			mob.sprite.texture:SetTexture(1,1,0,1);
		end
		if mob.hostile then
			if self.xml.scripts then
				mob.script = self.xml.scripts.default;
			end
		end
		mob.sprite.texture:SetSize(mob.width, mob.height);
		mob.sprite.texture:SetParent(self.frame);
		mob.sprite.obj = mob;
		mob.Turn = Gnomario_World_Turn;
		mob.Walk = Gnomario_World_Walk;
		tinsert(self.creatures, mob);
		self.physics:Put(mob, 0, 0);
	end
	return mob;
end

function Gnomario_World_Step(self, duration)
	self.physics:Step(duration);
	for n = 1, #self.sprites do
		local sprite = self.sprites[n];
		if sprite.id then
			sprite.texture:Show();
			sprite.texture:SetPoint("BOTTOMLEFT", self.frame, "BOTTOMLEFT", sprite.obj.x, sprite.obj.y);
		else
			sprite.texture:Hide();
		end
	end
end


function Gnomario_World_Create(frame,width,height,cb,db)
	local world = {};
	world.xml = db;
	world.frame = frame;
	world.environment = {};
	world.creatures = {};
	world.sprites = {};
	world.width = width;
	world.height = height;
	world.Clear = Gnomario_World_Unregister;
	world.Spawn_Creature = Gnomario_World_Spawn_Creature;
	world.Spawn_Object = Gnomario_World_Spawn_Object;
	world.physics = Gnomario_Physics_Chamber(width, height, cb);
	world.Step = Gnomario_World_Step;
	return world;
end
